"""
read and process temporal data from LS-DYNA .binout0000 files
author: Magdalena Heibeck

Input
---------
file *.binout0000   : temporal data from LS-DYNA simulations
parameters n_rotor1 and n_rotor2    : rotor speeds in rpm
file 0_casename.txt : specimen's name and initial position

Returns
------
file rotor_energies_over_time.csv       : rotor moments, speed, calculated power, energies
file numerical_energies_over_time.csv   : numerical energies
file tiebreak_contact_over_time.csv     : info on tiebreak contact (tied area and nodes)
"""

import numpy as np
import pandas as pd
from lasso.dyna import Binout
import sys


# manually define directory
#fn_sim = '...\\A050\\240416_A050-06'

# alternatively, set up script to access command-line arguments (to be called from a master script)
fn_sim = sys.argv[1]

# define rotor speeds in rpm
n_rotor1 = 21.7
n_rotor2 = 14

fn_results1 = fn_sim + '\\output\\rotor_energies_over_time.csv'
fn_results2 = fn_sim + '\\output\\numerical_energies_over_time.csv'
fn_results3 = fn_sim + '\\output\\tiebreak_contact_over_time.csv'

# read title
casename = open(fn_sim + '\\output\\0_casename.txt','r')
casename = casename.read()

# print info
print(f"running python-script: sim_binout_read_plate_2layers.py for case: {casename}")

# import data
binout = Binout(fn_sim + '\\*binout0000')
# binout.read()

## ROTOR MOMENTS AND ENERGY
# save xmoments (in Nmm) to array
xmoment = binout.read('bndout','velocity','rigidbodies','xmoment')
#ids = binout.read('bndout','velocity','rigidbodies','ids')

# create results dataframe from xmoments (there are also columns for rotor3 and rotor4 if present)
results = pd.DataFrame(xmoment[:,:2], columns=['moment_rotor1','moment_rotor2'])

# substitute weird values in first row by zeros
results.loc[0,'moment_rotor1'] = 0
results.loc[0,'moment_rotor2'] = 0

# convert moment units from Nmm to kNm and invert sign of rotor1
results['moment_rotor1'] = results['moment_rotor1'] * -1e-06
results['moment_rotor2'] = results['moment_rotor2'] * 1e-06

# insert casename in first position
results.insert(0,"case",casename)

# insert time in second position
time1 = binout.read('bndout','velocity','rigidbodies','time')
time1 = pd.DataFrame(time1, columns=['time'])
results.insert(1,"time",time1)

# add rotor speeds (in rpm)
results['n_rotor1'] = n_rotor1
results['n_rotor2'] = n_rotor2

# calculate mechanical power P (in kNm/s)
results['power_rotor1'] = 2 * np.pi * results['n_rotor1']/60 * results['moment_rotor1']
results['power_rotor2'] = 2 * np.pi * results['n_rotor2']/60 * results['moment_rotor2']

# calculate mechanical energy consumption W (in kNm)
results['energy_rotor1'] = np.sum(results['power_rotor1']*(results.loc[1,'time']-results.loc[0,'time']))
results['energy_rotor2'] = np.sum(results['power_rotor2']*(results.loc[1,'time']-results.loc[0,'time']))

# save results dataframe as .csv
results.to_csv(fn_results1, index=False)



## NUMERICAL ENERGIES
# start creating results2 df with time
time2 = binout.read('glstat','time')
results2 = pd.DataFrame(time2, columns=['time'])

# insert casename in first position
results2.insert(0,"case",casename)

# add data on energies to results df (in Nmm)
#binout.read('glstat')
results2['energy_internal'] = binout.as_df('glstat', 'internal_energy').values
results2['energy_kinetic'] = binout.as_df('glstat', 'kinetic_energy').values
results2['energy_hourglass'] = binout.as_df('glstat', 'hourglass_energy').values
results2['energy_sliding'] = binout.as_df('glstat', 'sliding_interface_energy').values
results2['energy_total'] = binout.as_df('glstat', 'total_energy').values
results2['energy_internal_eroded'] = binout.as_df('glstat', 'eroded_internal_energy').values
results2['energy_kinetic_eroded'] = binout.as_df('glstat', 'eroded_kinetic_energy').values
results2['energy_hourglass_eroded'] = binout.as_df('glstat', 'eroded_hourglass_energy').values
results2['energy_ratio'] = binout.as_df('glstat', 'energy_ratio').values
results2['energy_ratio_wo_eroded'] = binout.as_df('glstat', 'energy_ratio_wo_eroded').values
results2['added_mass'] = binout.as_df('glstat', 'added_mass').values * 1000*1000 # convert mass from t to g

# binout.read('matsum')
matsum_ids = binout.read('matsum', 'ids')
matsum_internal = binout.read('matsum', 'internal_energy')
matsum_internal = pd.DataFrame(matsum_internal, columns=matsum_ids)
matsum_hourglass = binout.read('matsum', 'hourglass_energy')
matsum_hourglass = pd.DataFrame(matsum_hourglass, columns=matsum_ids)
matsum_kinetic = binout.read('matsum', 'kinetic_energy')
matsum_kinetic = pd.DataFrame(matsum_kinetic, columns=matsum_ids)
matsum_added_mass = binout.read('matsum', 'mass') * 1000*1000 # convert mass from t to g
matsum_added_mass = pd.DataFrame(matsum_added_mass, columns=matsum_ids)

# overwrite energies from glstat (data looks too messy due to jumps in eroded energy of few elements)
results2['energy_internal'] = matsum_internal[3] + matsum_internal[4]
results2['energy_hourglass'] = matsum_hourglass[3] + matsum_hourglass[4]
results2['energy_kinetic'] = matsum_kinetic[3] + matsum_kinetic[4]
results2['added_mass'] = matsum_added_mass[3] + matsum_added_mass[4]
# add data for part-specific energies
results2['energy_internal_steel'] = matsum_internal[3]
results2['energy_internal_organo'] = matsum_internal[4]
results2['energy_hourglass_steel'] = matsum_hourglass[3]
results2['energy_hourglass_organo'] = matsum_hourglass[4]
results2['energy_kinetic_steel'] = matsum_kinetic[3]
results2['energy_kinetic_organo'] = matsum_kinetic[4]
results2['added_mass_steel'] = matsum_added_mass[3]
results2['added_mass_organo'] = matsum_added_mass[4]

# binout.read('sleout')
sleout_ids = binout.read('sleout', 'ids')
# binout.read('sleout', 'legend')
sleout_slidingenergy = binout.read('sleout', 'slave+master')
sleout_slidingenergy = pd.DataFrame(sleout_slidingenergy, columns=sleout_ids)
# add data for contact-specific sliding interface energies
results2['sleout_specimen_self'] = sleout_slidingenergy[1]
results2['sleout_specimen_to_shredder'] = sleout_slidingenergy[2]
results2['sleout_specimen_to_box'] = sleout_slidingenergy[3]
results2['sleout_steel_to_shredder'] = sleout_slidingenergy[4]
results2['sleout_organo_to_shredder'] = sleout_slidingenergy[5]
results2['sleout_tiebreak'] = sleout_slidingenergy[423]

# save results dataframe as .csv
results2.to_csv(fn_results2, index=False)


## TIEBREAK CONTACT
# extract tie_area and tie_count data for contacts 423(column 4) and optional 425(column 6) at first and last state from binout
time3 = binout.read('rcforc','time')
results3 = pd.DataFrame(time3, columns=['time'])

# check contact ids
# binout.read('rcforc', 'ids')
tie_area = binout.read('rcforc', 'tie_area')
tie_area = pd.DataFrame(tie_area)
tie_nodes = binout.read('rcforc', 'tie_count')
tie_nodes = pd.DataFrame(tie_nodes)

results3['contact_423_area'] = tie_area[8]
results3['contact_423_nodes'] = tie_nodes[8]


# insert casename in first position
results3.insert(0, "case", casename)

# delete first line (is zeros)
results3 = results3.drop(results3.iloc[[0]].index)

# save results dataframe as .csv
results3.to_csv(fn_results3, index=False)