"""
determine fragment properties (mass, composition, size) from .stl-files
author: Magdalena Heibeck

Input
---------
files *.stl                 : fragments to be analyzed (one file per material component and thickness)
file 0_casename.txt         : feed specimen's name
file .\ch4_profile_sim\3_fragment_characterization\stl_files\section_data_profile_for_python.csv: thickness and density of profile specimen's materials to calculate masses
or:  .\ch5_plates_sim\3_fragment_characterization\stl_files\section_data_plates_for_python.csv:   thickness and density of plate specimen's materials to calculate masses

Returns
------
file fragment_characterization.csv  : fragment-specific properties
                                        - mass (in g)   :
                                            - total mass
                                            - mass of steel, organosheet, and PA to calculate material composition of fragment
                                        - size (in mm)
                                            - x1, x2, x3: dimensions of oriented bounding box from principal component analysis (PCA)
                                            - d1, d2, d3: dimensions of minimum bounding box
files .\plots\*.png                 : contour plots of fragments aligned along first and second axes of PCA
"""


from paraview import servermanager as sm
from paraview import simple as pv
import os
import re
from paraview.vtk.numpy_interface import dataset_adapter as dsa
import numpy as np
import pandas as pd
import glob
import sys
from datetime import datetime
import compas.geometry
from matplotlib import pyplot as plt

# manually define directory
#fn_sim = '...\\221123_aframe_006_1-5mm_hor_000deg_ready'

# alternatively, set up script to access command-line arguments (to be called from a master script)
fn_sim = sys.argv[1]


# start run timer
start_time = datetime.now()
print(f"the script started at: {start_time}")

# define directories of .stl files and results file
fn_stl = fn_sim + '\\output\\stl'
fn_results = fn_sim + '\\output\\fragment_characterization.csv'

# read title
casename = open(fn_sim + '\\output\\0_casename.txt', 'r')
casename = casename.read()

# print info
print(f"running python-script: sim_characterize_fragments.py for case: {casename}")

# import part data with part names from simulation and corresponding material, thickness, and density
fn_part_data =  ".\ch4_profile_sim\3_fragment_characterization\stl_files\section_data_profile_for_python.csv"
#fn_part_data =  ".\ch5_plates_sim\3_fragment_characterization\stl_files\section_data_plates_for_python.csv"


part_data = pd.read_csv(fn_part_data, sep=';', skiprows=1)


# function definitions
def normal(triangles):
    "From https://stackoverflow.com/questions/71346322/numpy-area-of-triangle-and-equation-of-a-plane-on-which-triangle-lies-on"
    # The cross product of two sides is a normal vector
    return np.cross(triangles[:, 1] - triangles[:, 0],
                    triangles[:, 2] - triangles[:, 0], axis=1)


def get_tri_area(triangles):
    "From https://stackoverflow.com/questions/71346322/numpy-area-of-triangle-and-equation-of-a-plane-on-which-triangle-lies-on"
    # The norm of the cross product of two sides is twice the area
    return np.linalg.norm(normal(triangles), axis=1) / 2


# initialize dict for storing particle data of all regions
all_mats = {}
region_id_global = 0
# write name of all .stl files in list
stl_list = glob.glob(fn_stl + '\\*.stl')
## loop over list with .stl files
for i, fn in enumerate(stl_list):  # i=1
    # get only the filename from full directory
    # stl_file = fn.rpartition('\\')[2]
    stl_file = os.path.split(fn)[1]
    # load stl file into paraview
    stl_object = pv.STLReader(registrationName=stl_file, FileNames=[stl_list[i]])

    # compute connectivity. this creates the RegionId arrays
    connectivity1 = pv.Connectivity(registrationName='Connectivity1', Input=stl_object)
    connectivity1.RegionIdAssignmentMode = 'Cell Count Descending'

    # find out how many regions exist in the data
    vtk_data = sm.Fetch(connectivity1)
    # transform paraview object so something numpy like
    vtk_data = dsa.WrapDataObject(vtk_data)
    n_regions = len(np.unique(vtk_data.PointData[0]))
    print(f"{n_regions: >3} regions found in {stl_file}")

    # get coordinates of all regions
    all_coords = vtk_data.GetPoints()
    # store only part name (as in LS-Dyna)
    part_name = re.findall(r"final_state.(\S*).stl", stl_file)[0]

    ## for each region do stuff
    for region_id in range(n_regions):
        # get coords per region
        region_id_coords = np.array(all_coords[vtk_data.PointData[0] == region_id])

        # populate all_mats dict
        all_mats[region_id_global] = {}
        all_mats[region_id_global]['coords'] = region_id_coords
        all_mats[region_id_global]['visited'] = False  # flag for later on
        all_mats[region_id_global]['part_name'] = part_name
        all_mats[region_id_global]['part_of'] = None  # flag for later on
        # get material, thickness and density info from part_data object
        all_mats[region_id_global]['material'] = part_data['material'][part_data['part_name'] == part_name].values[0]
        part_thickness = part_data['thickness'][part_data['part_name'] == part_name].values[0]
        part_density = part_data['density'][part_data['part_name'] == part_name].values[0]
        all_mats[region_id_global]['thickness'] = part_thickness
        all_mats[region_id_global]['density'] = part_density

        # create a new 'Threshold'
        threshold1 = pv.Threshold(registrationName='Threshold1', Input=connectivity1)
        threshold1.Scalars = ['POINTS', 'RegionId']
        threshold1.LowerThreshold = region_id
        threshold1.UpperThreshold = region_id

        # compute triangle area for all triangles in this region
        thresh_data = sm.Fetch(threshold1)
        thresh_data = dsa.WrapDataObject(thresh_data)

        # len(thresh_data.PointData[0]) # n_points
        n_tris = len(thresh_data.CellData[0])
        # np.max(thresh_data.CellLocations)

        # reshape array to something useful
        triangle_number_list = np.reshape(thresh_data.Cells, (n_tris, 4))

        # remove first column. fist column is just '3'
        triangle_number_list = triangle_number_list[:, 1:]
        cell_coords = thresh_data.Points[triangle_number_list]
        # store part coordinates into all_mats dict
        all_mats[region_id_global]['cell_coords'] = cell_coords

        # test tri area function
        # a = np.array([[0,0,0],
        #               [0,-1,0],
        #               [-1,-1,0]])
        # cell_coords.shape
        # area(a[np.newaxis,:,:])

        # compute part area and mass
        part_area = get_tri_area(cell_coords).sum()
        all_mats[region_id_global]['area'] = part_area
        all_mats[region_id_global]['mass'] = part_area * part_density * part_thickness

        # finish loop
        region_id_global += 1

print(f"total amount of regions: {len(all_mats)}")
print(f"runtime until splitting with Paraview is finished (in h:mm:ss): {datetime.now() - start_time}")

# temporarily reduce size of dict all_mats:
# for i in range(100,len(all_mats)):
# print(i)
# del all_mats[i]

# initialise results dataframe
results = pd.DataFrame().from_dict({'fragment_id': [],
                                    'mass_steel': [],
                                    'mass_organo': [],
                                    'mass_PA': [],
                                    'coords': []
                                    }
                                   )


## function definition
# find out if two regions are actually one particle ("neighbours")
# can be neighbours even with a gap between them due to shell thicknesses (--> distance threshold)
def is_neigh(coords_loc, coords_to_test, dist_thres=4):
    #coords_loc = data.iloc[region_id]['coords'].copy()
    #coords_to_test = data.iloc[region_id_comp]['coords'].copy()
    # calculate distances between all nodes of two regions
    distances = [np.linalg.norm(coords_to_test - coords_loc[it, :], axis=1) for it in range(coords_loc.shape[0])]
    #distances = [np.sqrt(((coords_to_test - coords_loc[it, :])**2).sum(axis=1)) for it in range(coords_loc.shape[0])]
    # return TRUE/FALSE if particles are neighbours
    return np.any(np.array(distances) < dist_thres)


# initialize id for part_info
i = 1  # ?
## loop over all regions
for region_id in all_mats:
    if not all_mats[region_id]['visited']:
        # initialize new part
        part_info = {'fragment_id': [i],
                     'mass_steel': [0],
                     'mass_organo': [0],
                     'mass_PA': [0],
                     'coords': [np.empty((0, 3))]
                     }
        i += 1
        # set flag
        all_mats[region_id]['visited'] = True
        # store coordinates
        coords = all_mats[region_id]['coords']

        # add this mass to correct material mass
        region = all_mats[region_id]
        summed_mass = part_info[f"mass_{region['material']}"][0] + region['mass']
        part_info[f"mass_{region['material']}"] = [summed_mass]
        # concatenate coordinates
        part_info['coords'] = [np.vstack((part_info['coords'][0], region['coords']))]

        ## loop again over all regions to check if two regions are neighbours
        for region_id_comp in all_mats:
            region = all_mats[region_id_comp]
            if not region['visited']:

                coords_to_test = region['coords']
                # calculate distances between all nodes of two regions
                distances = [np.linalg.norm(coords_to_test - coords[i, :], axis=1) for i in
                             range(coords.shape[0])]
                # distance threshold: can be neighbours even with a gap between them due to shell thicknesses
                dist_thres = 4
                # check if two regions are neighbours
                if np.any(np.array(distances) < dist_thres):
                    # print(f"{region_id_comp}: {np.sum(np.array(distances) < dist_thres)}")
                    region['visited'] = True
                    # add masses for specific materials and concatenate coordinates
                    summed_mass = part_info[f"mass_{region['material']}"][0] + region['mass']
                    part_info[f"mass_{region['material']}"] = [summed_mass]
                    part_info['coords'] = [np.vstack((part_info['coords'][0], region['coords']))]
        results = pd.concat((results, pd.DataFrame().from_dict(part_info)))

# print total material masses
print(f"mass steel: {np.round(results['mass_steel'].sum())}")
print(f"mass organo: {np.round(results['mass_organo'].sum())}")
print(f"mass PA: {np.round(results['mass_PA'].sum())}")

## sum up neighbouring particles until no new neighbours are found
# set flag
neigh_found = True

# do things until no neighbouring particles are found
while neigh_found is not False:
    # initialise temp dataframe
    results_tmp = pd.DataFrame()
    neigh_found = False

    # compare all particles with each other
    for i in range(results.shape[0]):
        for j in range(results.shape[0]):
            if i == j:
                # don't compare particle with itself
                continue
            print(f"testing now: i={i} and j={j}")
            if is_neigh(results.iloc[i]['coords'].copy(), results.iloc[j]['coords'].copy(), dist_thres=4) and i != j:
                print("Found neighbor")
                print(f"i={i} and j={j}")
                # sum masses
                mass_pa = results.iloc[i]['mass_PA'] + results.iloc[j]['mass_PA']
                mass_organo = results.iloc[i]['mass_organo'] + results.iloc[j]['mass_organo']
                mass_steel = results.iloc[i]['mass_steel'] + results.iloc[j]['mass_steel']
                # concatenate coordinates
                coords = [np.vstack((results.iloc[i]['coords'], results.iloc[j]['coords']))]
                # populate temp results dataframe with new merged particle
                results_tmp = pd.concat((results_tmp, pd.DataFrame({
                    'fragment_id': i,
                    'mass_PA': mass_pa,
                    'mass_organo': mass_organo,
                    'mass_steel': mass_steel,
                    'coords': coords
                })))
                # reset flag
                neigh_found = (i, j)

                # populate temp results dataframe with remaining unmerged particles
                for fill_i in range(results.shape[0]):
                    if fill_i not in neigh_found:
                        results_tmp = pd.concat((results_tmp, results.iloc[[fill_i]]))
                break
        # break loop after exactly one neighbour pair is found
        if neigh_found is not False:
            break

    if neigh_found is not False:
        # overwrite results dataframe and repeat loop until no new neighbour is found
        results = results_tmp
print("no new neighbours have been found")
print(f"runtime until neighbouring regions are merged (in h:mm:ss): {datetime.now() - start_time}")

# reset df rowindex to monotonic increasing (was just zeros before)
results.index = results.reset_index().index

## get bounding box for each particle
for part in range(results.shape[0]):  # part = 1
    coords = results.coords[part]
    # coords = results.iloc[[0]].coords.values[part] #hier gibt es das Problem:  IndexError: index 1 is out of bounds for axis 0 with size 1
    coords = pd.DataFrame(coords, columns=['0', '1', '2'])
    # coords = pd.DataFrame().from_dict(all_mats[0]["coords"])

    # Calculate means
    means = np.mean(coords, axis=0)
    # Center the data (means == 0)
    coords_centered = coords - means

    # singular value decomposition 3D
    # U: unitary matrix
    # S: diagonal matrix of singular values
    # V: columns of V are principal directions/axes
    U, s, Vt = np.linalg.svd(coords_centered, full_matrices=False)
    V = Vt.T
    S = np.diag(s)
    # principal components (columns of US) = projections of the data on the principal axes: principal_components = U.dot(S)
    # Rotate the data i.e. align the eigen vector to the cartesian basis
    coords_aligned = np.matmul(coords, V)
    # center data (for plotting)
    means = np.mean(coords_aligned, axis=0)
    coords_aligned = coords_aligned - means

    # calculate dimensions of oriented bounding box
    len_axis0 = np.max(coords_aligned[0]) - np.min(coords_aligned[0])
    len_axis1 = np.max(coords_aligned[1]) - np.min(coords_aligned[1])
    len_axis2 = np.max(coords_aligned[2]) - np.min(coords_aligned[2])

    # write dimensions to results dataframe
    results.loc[part, 'x_1'] = len_axis0
    results.loc[part, 'x_2'] = len_axis1
    results.loc[part, 'x_3'] = len_axis2

    # Convert coords_aligned to a NumPy array
    coords_aligned_np = np.array(coords_aligned)

    # Compute the oriented minimum bounding box of a set of points in 3D space
    bbox = compas.geometry.oriented_bounding_box_numpy(coords_aligned_np)
    a = compas.geometry.length_vector(compas.geometry.subtract_vectors(bbox[1], bbox[0]))
    b = compas.geometry.length_vector(compas.geometry.subtract_vectors(bbox[3], bbox[0]))
    c = compas.geometry.length_vector(compas.geometry.subtract_vectors(bbox[4], bbox[0]))

    # sort dimensions of minimum bounding box
    minbox_1 = max(a, b, c)
    minbox_3 = min(a, b, c)
    if (minbox_1 > a > minbox_3): minbox_2 = a
    elif (minbox_1 > b > minbox_3): minbox_2 = b
    else: minbox_2 = c

    # write dimensions of minimum bounding box to results dataframe
    results.loc[part, 'd_1'] = minbox_1
    results.loc[part, 'd_2'] = minbox_2
    results.loc[part, 'd_3'] = minbox_3

    # plot: set title
    if results.loc[part, 'mass_steel'] == 0:
        temp_title = f'{casename}: Part {part + 1} (liberated FRP)'
    elif (results.loc[part, 'mass_organo'] == 0) & (results.loc[part, 'mass_PA'] == 0):
        temp_title = f'{casename}: Part {part + 1} (liberated steel)'
    else:
        temp_title = f'{casename}: Part {part + 1} (unliberated)'

    # plot individual fragment
    plt.figure()  # Create a new figure for the individual plot
    plt.scatter(coords_aligned[0], coords_aligned[1], c='b', linewidths=1)  # Plot the part
    plt.xlabel('X in mm (1st main dimension)')
    plt.ylabel('Y in mm (2nd main dimension)')
    plt.title(temp_title + f', d_2 = {minbox_2:.1f} mm')  # Set title
    plt.gca().set_aspect('equal', adjustable='box')  # Set equal aspect ratio
    plt.savefig(fn_sim + f'\\output\\plots\\fragments_aligned_part{part + 1}.png')  # Save the individual plot as a PNG file
    plt.close()  # Close the figure to release memory

print(f"runtime until bounding boxes are computed (in h:mm:ss): {datetime.now() - start_time}")

"""
# plot
from matplotlib import pyplot as plt

plt.scatter(coords[1],coords[0])
plt.scatter(coords[2],coords[0])
plt.scatter(coords[1],coords[2])

fig = plt.figure(figsize=plt.figaspect(1)*2)
ax = fig.add_subplot(projection='3d', proj_type='ortho')
ax.set_xlim3d(-145, 145)
ax.set_ylim3d(-145, 145)
ax.set_zlim3d(-145, 145)
ax.scatter3D(coords[0], coords[1], coords[2])
ax.scatter3D(coords_centered[0], coords_centered[1], coords_centered[2])
ax.scatter3D(coords_aligned[0], coords_aligned[1], coords_aligned[2])
plt.show()
"""

# POST-Processing / cleaning of df
# delete coords column from results to save space
del results['coords']
# sort results dataframe by axis0 in descending order
results = results.sort_values(by=['d_2'], ascending=False)
# reset fragment_id (there were jumps before)
results.fragment_id = results.reset_index().index
results.fragment_id = results.fragment_id +1

# add column for casename in first position
results.insert(0, "case", casename)
# add column for total mass in third position
results.insert(2, "mass", results['mass_steel'] + results['mass_organo'] + results['mass_PA'])
# results['mass'] = results['mass_steel']+results['mass_organo']+results['mass_PA']


# transform results dict to pd dataframe and save as .csv
pd.DataFrame(results).to_csv(fn_results, index=False)